/*
 * this file is a part of amp software, (C) tomislav uzelac 1996,1997
 */

/*
 * getbits.c  bit level routines, input buffer
 * 
 * Created by: tomislav uzelac  Apr 1996 better synchronization, tomislav
 * uzelac, Apr 23 1997
 */
/*
 * Changed into a library to play back mp3s for Junkie FTP.
 */
#include <jamp.h>
#include <amp.h>
#include <audio.h>
#include <string.h>

#define	GETBITS
#include <getbits.h>

/*
 * buffer and bit manipulation functions ***************************************
 */
static inline int
_fillbfr(amp_file *jamp, unsigned int size)
{
	jamp->_bptr = 0;
	return get_input(jamp, jamp->_buffer, size);
}

static inline int
readsync(amp_file *jamp)
{
	jamp->_bptr = 0;
	jamp->_buffer[0] = jamp->_buffer[1];
	jamp->_buffer[1] = jamp->_buffer[2];
	jamp->_buffer[2] = jamp->_buffer[3];
	return get_input(jamp, &jamp->_buffer[3], 1);
}

static inline unsigned int
_getbits(amp_file *jamp, int n)
{
	unsigned int    pos, ret_value;

	pos = jamp->_bptr >> 3;
	ret_value = jamp->_buffer[pos] << 24 |
		jamp->_buffer[pos + 1] << 16 |
		jamp->_buffer[pos + 2] << 8 |
		jamp->_buffer[pos + 3];
	ret_value <<= jamp->_bptr & 7;
	ret_value >>= 32 - n;
	jamp->_bptr += n;
	return ret_value;
}

int
fillbfr(amp_file *jamp, unsigned int advance)
{
	int             overflow, retval;
	int append;

	append = jamp_get_append(jamp);
	retval = get_input(jamp, &jamp->buffer[append], advance);

	if (append + advance >= BUFFER_SIZE) {
		overflow = append + advance - BUFFER_SIZE;
		memcpy(jamp->buffer, &jamp->buffer[BUFFER_SIZE], overflow);
		if (overflow < 4)
			memcpy(&jamp->buffer[BUFFER_SIZE], jamp->buffer, 4);
		append = overflow;
	} else {
		if (append == 0)
			memcpy(&jamp->buffer[BUFFER_SIZE], jamp->buffer, 4);
		append += advance;
	}
	jamp_set_append(jamp, append);

	return retval;
}

unsigned int
getbits(amp_file *jamp, int n)
{
	if (n && jamp) {
		unsigned int    pos, ret_value;

		pos = jamp->data >> 3;
		ret_value = jamp->buffer[pos] << 24 |
			jamp->buffer[pos + 1] << 16 |
			jamp->buffer[pos + 2] << 8 |
			jamp->buffer[pos + 3];
		ret_value <<= jamp->data & 7;
		ret_value >>= 32 - n;

		jamp->data += n;
		jamp->data &= (8 * BUFFER_SIZE) - 1;

		return ret_value;
	} else
		return 0;
}

/*
 * header and side info parsing stuff ******************************************
 */
static inline void
parse_header(amp_file *jamp, struct AUDIO_HEADER * header)
{
	header->ID = _getbits(jamp, 1);
	header->layer = _getbits(jamp, 2);
	header->protection_bit = _getbits(jamp, 1);
	header->bitrate_index = _getbits(jamp, 4);
	header->sampling_frequency = _getbits(jamp, 2);
	header->padding_bit = _getbits(jamp, 1);
	header->private_bit = _getbits(jamp, 1);
	header->mode = _getbits(jamp, 2);
	header->mode_extension = _getbits(jamp, 2);
	if (!header->mode)
		header->mode_extension = 0;
	header->copyright = _getbits(jamp, 1);
	header->original = _getbits(jamp, 1);
	header->emphasis = _getbits(jamp, 2);
}

static inline int
header_sanity_check(struct AUDIO_HEADER * header)
{
	if (header->layer == 0 ||
	    header->bitrate_index == 15 ||
	    header->sampling_frequency == 3)
		return -1;

	/*
	 * an additional check to make shure that stuffing never gets
	 * mistaken for a syncword. This rules out some legal layer1 streams,
	 * but who cares about layer1 anyway :-). I must get this right
	 * sometime.
	 */
	if (header->ID == 1 && header->layer == 3 && header->protection_bit == 1)
		return -1;
	return 0;
}


int
gethdr(amp_file *jamp, struct AUDIO_HEADER * header)
{
	int             s, retval;
	struct AUDIO_HEADER tmp;

	/*
	 * TODO: add a simple byte counter to check only first, say, 1024
	 * bytes for a new header and then return GETHDR_SYN
	 */
	if ((retval = _fillbfr(jamp, 4)) != 0)
		return retval;

	for (;;) {
		while ((s = _getbits(jamp, 12)) != 0xfff) {
			if (s == 0xffe) {
				parse_header(jamp, &tmp);
				if (header_sanity_check(&tmp) == 0)
					return GETHDR_NS;
			}
			if ((retval = readsync(jamp)) != 0)
				return retval;
		}

		parse_header(jamp, &tmp);
		if (header_sanity_check(&tmp) != 0) {
			if ((retval = readsync(jamp)) != 0)
				return retval;
		} else
			break;
	}

	if (tmp.layer == 3)
		return GETHDR_FL1;
	/* if (tmp.layer==2) return GETHDR_FL2; */
	if (tmp.bitrate_index == 0)
		return GETHDR_FF;

	memcpy(header, &tmp, sizeof(tmp));
	return 0;
}

/*
 * dummy function, to get crc out of the way
 */
void
getcrc(amp_file *jamp)
{
	_fillbfr(jamp, 2);
	_getbits(jamp, 16);
}

/*
 * sizes of side_info: MPEG1   1ch 17    2ch 32 MPEG2   1ch  9    2ch 17
 */
void
getinfo(amp_file *jamp, struct AUDIO_HEADER * header, struct SIDE_INFO * info)
{
	int             gr, ch, scfsi_band, region, window;
	int             nch;

	if (header->mode == 3) {
		nch = 1;
		if (header->ID) {
			_fillbfr(jamp, 17);
			info->main_data_begin = _getbits(jamp, 9);
			_getbits(jamp, 5);
		} else {
			_fillbfr(jamp, 9);
			info->main_data_begin = _getbits(jamp, 8);
			_getbits(jamp, 1);
		}
	} else {
		nch = 2;
		if (header->ID) {
			_fillbfr(jamp, 32);
			info->main_data_begin = _getbits(jamp, 9);
			_getbits(jamp, 3);
		} else {
			_fillbfr(jamp, 17);
			info->main_data_begin = _getbits(jamp, 8);
			_getbits(jamp, 2);
		}
	}

	if (header->ID)
		for (ch = 0; ch < nch; ch++)
			for (scfsi_band = 0; scfsi_band < 4; scfsi_band++)
				info->scfsi[ch][scfsi_band] = _getbits(jamp, 1);

	for (gr = 0; gr < (header->ID ? 2 : 1); gr++)
		for (ch = 0; ch < nch; ch++) {
			info->part2_3_length[gr][ch] = _getbits(jamp, 12);
			info->big_values[gr][ch] = _getbits(jamp, 9);
			info->global_gain[gr][ch] = _getbits(jamp, 8);
			if (header->ID)
				info->scalefac_compress[gr][ch] = _getbits(jamp, 4);
			else
				info->scalefac_compress[gr][ch] = _getbits(jamp, 9);
			info->window_switching_flag[gr][ch] = _getbits(jamp, 1);

			if (info->window_switching_flag[gr][ch]) {
				info->block_type[gr][ch] = _getbits(jamp, 2);
				info->mixed_block_flag[gr][ch] = _getbits(jamp, 1);

				for (region = 0; region < 2; region++)
					info->table_select[gr][ch][region] = _getbits(jamp, 5);
				info->table_select[gr][ch][2] = 0;

				for (window = 0; window < 3; window++)
					info->subblock_gain[gr][ch][window] = _getbits(jamp, 3);
			} else {
				for (region = 0; region < 3; region++)
					info->table_select[gr][ch][region] = _getbits(jamp, 5);

				info->region0_count[gr][ch] = _getbits(jamp, 4);
				info->region1_count[gr][ch] = _getbits(jamp, 3);
				info->block_type[gr][ch] = 0;
			}

			if (header->ID)
				info->preflag[gr][ch] = _getbits(jamp, 1);
			info->scalefac_scale[gr][ch] = _getbits(jamp, 1);
			info->count1table_select[gr][ch] = _getbits(jamp, 1);
		}
	return;
}

int
dummy_getinfo(amp_file *jamp, int n)
{
	n -= 4;
	if (fseek(jamp->file, n, SEEK_CUR) != 0) {
		if (feof(jamp->file))
			return GETHDR_EOF;
		else
			return GETHDR_ERR;
	}
	return 0;
}

int
rewind_stream(amp_file *jamp, int nbytes)
{
	nbytes += 5;
	if (fseek(jamp->file, -nbytes, SEEK_CUR) != 0) {
		/* what if we need to be at the very beginning? */
		nbytes--;
		if (fseek(jamp->file, -nbytes, SEEK_CUR) != 0)
			return GETHDR_ERR;
	}
	return 0;
}

static inline int
get_input(amp_file *jamp, unsigned char *bp, unsigned int size)
{
	if (fread(bp, 1, size, jamp->file) != size) {
		if (feof(jamp->file))
			return GETHDR_EOF;
		else
			return GETHDR_ERR;
	}
	return 0;
}
