/*
 * this file is a part of amp software, (C) tomislav uzelac 1996,1997
 */

/*
 * audio.c	main amp source file
 * 
 * Created by: tomislav uzelac	Apr 1996 Karl Anders Oygard added the IRIX
 * code, 10 Mar 1997. Ilkka Karvinen fixed /dev/dsp initialization, 11 Mar
 * 1997. Lutz Vieweg added the HP/UX code, 14 Mar 1997. Dan Nelson added
 * FreeBSD modifications, 23 Mar 1997. Andrew Richards complete
 * reorganisation, new features, 25 Mar 1997 Edouard Lafargue added sajber
 * jukebox support, 12 May 1997
 */

/*
 * Changed into a library to play back mp3s for Junkie FTP.
 */
#include <jamp.h>
#include <amp.h>

#include <sys/types.h>
#include <sys/stat.h>

#ifndef __BEOS__
#include <sys/uio.h>
#endif

#include <sys/socket.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#define AUDIO
#include <transform.h>
#include <audio.h>
#include <getbits.h>
#include <huffman.h>
#include <layer2.h>
#include <layer3.h>
#include <misc2.h>

#ifndef __BEOS__
typedef int     bool;
#endif


int
write_audio(amp_file *jamp, char *buf)
{
	int j;

	/* playback the audio? stored in sample_buffer. */
	
	j = 32 * 18 * ((jamp->info.channels == 2) ? 2 : 1) * sizeof(short);
	memcpy(buf, jamp->sample_buffer, j);
	
	return j;
}



void
jamp_setup()
{
	initialise_decoder();
	return;
}

int
jamp_open(FILE *f, amp_file *jamp)
{
	struct AUDIO_HEADER header;
	int             g;

	if (!f || !jamp)
		return -1;

	jamp->file = f;
	initialise_globals(jamp);	/* ?? XXX remove globals.. */

	/* get the first valid header. */
	if ((g = gethdr(jamp, &header)) != 0) {
		// report_header_error(g);
		return -1;
	}

	/* if the protection bit is not set, there is a CRC check. */
	if (header.protection_bit == 0)
		getcrc(jamp);
		
	/* basically some error checking? */
	if (header.layer == 1) {
		if (layer3_frame(jamp, NULL, &header)) {
			// warn(" error. blip.\n");
			return -1;
		}
	} else if (header.layer == 2) {
		if (layer2_frame(jamp, NULL, &header)) {
			// warn(" error. blip.\n");
			return -1;
		}
	}

	return 1;
}

char *
jamp_read(amp_file *jamp, int *size)
{

	struct AUDIO_HEADER header;
	int             cnt, g, snd_eof;
	int total, j;
	char *pnt = NULL;

	/*
	 * decoder loop **********************************
	 */
	memset(&header, 0, sizeof(struct AUDIO_HEADER));
	snd_eof = 0;
	cnt = 0;
	total = 0;
	if ((g = gethdr(jamp, &header)) != 0) {
		/* invalid header. */
		// report_header_error(g);
		snd_eof = 1;
		cnt = -1;
		return NULL;
	}

	j = 32 * 18 * ((jamp->info.channels == 2) ? 2 : 1) * sizeof(short);
	if (header.layer == 1)
		j *= jamp->info.channels;
	else if (header.layer == 2)
		j *= 2;
	

	/* allocate memory for the data.. */
	if (!(pnt = malloc(j))) 
		return NULL;
	
	if (size) 
		*size = j;

	if (header.protection_bit == 0)
		getcrc(jamp);	/* file has got a crc check. */
			
	if (header.layer == 1) {
		if (layer3_frame(jamp, pnt, &header)) {
			// warn(" error. blip.\n");
			return NULL;
		}
	} else if (header.layer == 2) {
		if (layer2_frame(jamp, pnt, &header)) {
			// warn(" error. blip.\n");
			return NULL;
		}
	}
	

	return pnt;
}

int
jamp_set_channels(amp_file *jamp, int chans)
{
	if (!jamp)
		return 0;
	
	jamp->info.channels = chans;
	return 1;
}

int
jamp_get_channels(amp_file *jamp)
{
	if (!jamp)
		return 0;
	
	return jamp->info.channels;
}

int
jamp_set_append(amp_file *jamp, int append)
{
	if (!jamp)
		return 0;
	
	jamp->append = append;
	return 1;
}

int
jamp_get_append(amp_file *jamp)
{
	if (!jamp)
		return 0;
		
	return jamp->append;
}

int
jamp_get_dirty(amp_file *jamp)
{
	return ((!jamp) ? (0) : (jamp->f_bdirty));
}

int
jamp_set_dirty(amp_file *jamp, int bdirty)
{
	if (!jamp)
		return 0;

	jamp->f_bdirty = bdirty;
	return 1;
}

int
jamp_get_bclean_bytes(amp_file *jamp)
{
	return ((!jamp) ? (0) : (jamp->bclean_bytes));
}

int
jamp_set_bclean_bytes(amp_file *jamp, int bclean)
{
	if (!jamp)
		return 0;

	jamp->bclean_bytes = bclean;
	return 1;
}



/*
 * call this once at the beginning
 */
void 
initialise_decoder(void)
{
	premultiply();
	imdct_init();
	calculate_t43();
}

/*
 * call this before each file is played
 */
void 
initialise_globals(amp_file *jamp)
{
	/* nch, append, data = 0*/
	jamp_set_dirty(jamp, TRUE);
	jamp_set_bclean_bytes(jamp, 0);

	memset(jamp->s, 0, sizeof(jamp->s));
	memset(jamp->res, 0, sizeof(jamp->res));
}

#if 0
void 
report_header_error(int err)
{
	switch (err) {
		case GETHDR_ERR:die("error reading mpeg bitstream. exiting.\n");
		break;
	case GETHDR_NS:
		warn("this is a file in MPEG 2.5 format, which is not defined\n");
		warn("by ISO/MPEG. It is \"a special Fraunhofer format\".\n");
		warn("amp does not support this format. sorry.\n");
		break;
	case GETHDR_FL1:
		warn("ISO/MPEG layer 1 is not supported by amp (yet).\n");
		break;
	case GETHDR_FF:
		warn("free format bitstreams are not supported. sorry.\n");
		break;
	case GETHDR_SYN:
		warn("oops, we're out of sync.\n");
		break;
	case GETHDR_EOF:
	default:;		/* some stupid compilers need the semicolon */
	}
}
#endif


